@@ -39,7 +39,7 @@
  *
  * @{
  */
-typedef struct RofiViewState   RofiViewState;
+typedef struct RofiViewState RofiViewState;
 typedef enum
 {
     /** Create a menu for entering text */
@@ -52,8 +52,6 @@ typedef enum
     MENU_ERROR_DIALOG  = 4,
     /** INDICATOR */
     MENU_INDICATOR     = 8,
-    /** Show column */
-    MENU_PROMPT_COLON  = 16,
 } MenuFlags;
 
 /**
