@@ -31,6 +31,7 @@
 #ifdef ENABLE_DRUN
 #include <stdlib.h>
 #include <stdio.h>
+#include <limits.h>
 
 #include <unistd.h>
 #include <limits.h>
@@ -52,7 +53,9 @@
 #include "nkutils-xdg-theme.h"
 #include "xcb.h"
 
-#define DRUN_CACHE_FILE    "rofi2.druncache"
+#define DRUN_CACHE_FILE    "rofi3.druncache"
+
+#define DRUN_GROUP_NAME    "Desktop Entry"
 
 /**
  * Store extra information about the entry.
@@ -66,6 +69,8 @@ typedef struct
     char *path;
     /* Application id (.desktop filename) */
     char *app_id;
+    /* Desktop id */
+    char *desktop_id;
     /* Icon stuff */
     char *icon_name;
     /* Icon size is used to indicate what size is requested by the gui.
@@ -80,27 +85,57 @@ typedef struct
     char            *name;
     /* Generic Name */
     char            *generic_name;
+    /* Categories */
     char            **categories;
+    /* Comments */
+    char            *comment;
 
     GKeyFile        *key_file;
+
+    gint            sort_index;
 } DRunModeEntry;
 
+typedef struct
+{
+    const char *entry_field_name;
+    gboolean   enabled;
+} DRunEntryField;
+
+typedef enum
+{
+    DRUN_MATCH_FIELD_NAME,
+    DRUN_MATCH_FIELD_GENERIC,
+    DRUN_MATCH_FIELD_EXEC,
+    DRUN_MATCH_FIELD_CATEGORIES,
+    DRUN_MATCH_FIELD_COMMENT,
+    DRUN_MATCH_NUM_FIELDS,
+} DRunMatchingFields;
+
+static DRunEntryField matching_entry_fields[DRUN_MATCH_NUM_FIELDS] = {
+    { .entry_field_name = "name",       .enabled = TRUE,  },
+    { .entry_field_name = "generic",    .enabled = TRUE,  },
+    { .entry_field_name = "exec",       .enabled = TRUE,  },
+    { .entry_field_name = "categories", .enabled = TRUE,  },
+    { .entry_field_name = "comment",    .enabled = FALSE, }
+};
+
 typedef struct
 {
     NkXdgThemeContext *xdg_context;
     DRunModeEntry     *entry_list;
     unsigned int      cmd_list_length;
     unsigned int      cmd_list_length_actual;
-    unsigned int      history_length;
     // List of disabled entries.
     GHashTable        *disabled_entries;
     unsigned int      disabled_entries_length;
     GThreadPool       *pool;
-
     unsigned int      expected_line_height;
     DRunModeEntry     quit_entry;
+
     // Theme
     const gchar       *icon_theme;
+    // DE
+    gchar             **current_desktop_list;
 } DRunModePrivateData;
 
 struct RegexEvalArg
@@ -108,6 +143,7 @@ struct RegexEvalArg
     DRunModeEntry *e;
     gboolean      success;
 };
+
 static gboolean drun_helper_eval_cb ( const GMatchInfo *info, GString *res, gpointer data )
 {
     // TODO quoting is not right? Find description not very clear, need to check.
@@ -185,7 +221,7 @@ static void exec_cmd_entry ( DRunModeEntry *e )
     }
 
     const gchar *fp        = g_strstrip ( str );
-    gchar       *exec_path = g_key_file_get_string ( e->key_file, "Desktop Entry", "Path", NULL );
+    gchar       *exec_path = g_key_file_get_string ( e->key_file, DRUN_GROUP_NAME, "Path", NULL );
     if ( exec_path != NULL && strlen ( exec_path ) == 0 ) {
         // If it is empty, ignore this property. (#529)
         g_free ( exec_path );
@@ -197,19 +233,18 @@ static void exec_cmd_entry ( DRunModeEntry *e )
         .icon   = e->icon_name,
         .app_id = e->app_id,
     };
-    gboolean                 sn       = g_key_file_get_boolean ( e->key_file, "Desktop Entry", "StartupNotify", NULL );
+    gboolean                 sn       = g_key_file_get_boolean ( e->key_file, DRUN_GROUP_NAME, "StartupNotify", NULL );
     gchar                    *wmclass = NULL;
-    if ( sn && g_key_file_has_key ( e->key_file, "Desktop Entry", "StartupWMClass", NULL ) ) {
-        context.wmclass = wmclass = g_key_file_get_string ( e->key_file, "Desktop Entry", "StartupWMClass", NULL );
+    if ( sn && g_key_file_has_key ( e->key_file, DRUN_GROUP_NAME, "StartupWMClass", NULL ) ) {
+        context.wmclass = wmclass = g_key_file_get_string ( e->key_file, DRUN_GROUP_NAME, "StartupWMClass", NULL );
     }
 
     // Returns false if not found, if key not found, we don't want run in terminal.
-    gboolean terminal = g_key_file_get_boolean ( e->key_file, "Desktop Entry", "Terminal", NULL );
+    gboolean terminal = g_key_file_get_boolean ( e->key_file, DRUN_GROUP_NAME, "Terminal", NULL );
     if ( helper_execute_command ( exec_path, fp, terminal, sn ? &context : NULL ) ) {
         char *path = g_build_filename ( cache_dir, DRUN_CACHE_FILE, NULL );
-        char *key  = g_strdup_printf ( "%s:::%s", e->root, e->path );
-        history_set ( path, key );
-        g_free ( key );
+        // Store it based on the unique identifiers (desktop_id).
+        history_set ( path, e->desktop_id );
         g_free ( path );
     }
     g_free ( wmclass );
@@ -234,29 +269,29 @@ static gboolean read_desktop_file ( DRunModePrivateData *pd, const char *root, c
 
     // Check if item is on disabled list.
     if ( g_hash_table_contains ( pd->disabled_entries, id ) ) {
-        g_debug ( "Skipping: %s, was previously seen.", id );
+        g_debug ( "[%s] [%s] Skipping, was previously seen.", id, path );
         return TRUE;
     }
     GKeyFile *kf    = g_key_file_new ();
     GError   *error = NULL;
-    g_key_file_load_from_file ( kf, path, 0, &error );
+    gboolean res    = g_key_file_load_from_file ( kf, path, 0, &error );
     // If error, skip to next entry
-    if ( error != NULL ) {
-        g_debug ( "Failed to parse desktop file: %s because: %s", path, error->message );
+    if ( !res ) {
+        g_debug ( "[%s] [%s] Failed to parse desktop file because: %s.", id, path, error->message );
         g_error_free ( error );
         g_key_file_free ( kf );
         return FALSE;
     }
     // Skip non Application entries.
-    gchar *key = g_key_file_get_string ( kf, "Desktop Entry", "Type", NULL );
+    gchar *key = g_key_file_get_string ( kf, DRUN_GROUP_NAME, "Type", NULL );
     if ( key == NULL ) {
         // No type? ignore.
-        g_debug ( "Skipping desktop file: %s because: No type indicated", path );
+        g_debug ( "[%s] [%s] Invalid desktop file: No type indicated", id, path );
         g_key_file_free ( kf );
         return FALSE;
     }
     if ( g_strcmp0 ( key, "Application" ) ) {
-        g_debug ( "Skipping desktop file: %s because: Not of type application (%s)", path, key );
+        g_debug ( "[%s] [%s] Skipping desktop file: Not of type application (%s)", id, path, key );
         g_free ( key );
         g_key_file_free ( kf );
         return FALSE;
@@ -264,35 +299,71 @@ static gboolean read_desktop_file ( DRunModePrivateData *pd, const char *root, c
     g_free ( key );
 
     // Name key is required.
-    if ( !g_key_file_has_key ( kf, "Desktop Entry", "Name", NULL ) ) {
-        g_debug ( "Invalid DesktopFile: '%s', no 'Name' key present.", path );
+    if ( !g_key_file_has_key ( kf, DRUN_GROUP_NAME, "Name", NULL ) ) {
+        g_debug ( "[%s] [%s] Invalid desktop file: no 'Name' key present.", id, path );
         g_key_file_free ( kf );
         return FALSE;
     }
 
     // Skip hidden entries.
-    if ( g_key_file_get_boolean ( kf, "Desktop Entry", "Hidden", NULL ) ) {
-        g_debug ( "Adding desktop file: %s to disabled list because: Hdden", path );
+    if ( g_key_file_get_boolean ( kf, DRUN_GROUP_NAME, "Hidden", NULL ) ) {
+        g_debug ( "[%s] [%s] Adding desktop file to disabled list: 'Hidden' key is true", id, path );
         g_key_file_free ( kf );
         g_hash_table_add ( pd->disabled_entries, g_strdup ( id ) );
         return FALSE;
     }
+    if ( pd->current_desktop_list ) {
+        gboolean show = TRUE;
+        // If the DE is set, check the keys.
+        if ( g_key_file_has_key ( kf, DRUN_GROUP_NAME, "OnlyShowIn", NULL ) ) {
+            gsize llength = 0;
+            show = FALSE;
+            gchar **list = g_key_file_get_string_list ( kf, DRUN_GROUP_NAME, "OnlyShowIn", &llength, NULL );
+            if ( list ) {
+                for ( gsize lcd = 0; !show && pd->current_desktop_list[lcd]; lcd++ ) {
+                    for ( gsize lle = 0; !show && lle < llength; lle++ ) {
+                        show = ( g_strcmp0  ( pd->current_desktop_list[lcd], list[lle] ) == 0 );
+                    }
+                }
+                g_strfreev ( list );
+            }
+        }
+        if ( show && g_key_file_has_key ( kf, DRUN_GROUP_NAME, "NotShowIn", NULL ) ) {
+            gsize llength = 0;
+            gchar **list  = g_key_file_get_string_list ( kf, DRUN_GROUP_NAME, "NotShowIn", &llength, NULL );
+            if ( list ) {
+                for ( gsize lcd = 0; show && pd->current_desktop_list[lcd]; lcd++ ) {
+                    for ( gsize lle = 0; show && lle < llength; lle++ ) {
+                        show = !( g_strcmp0  ( pd->current_desktop_list[lcd], list[lle] ) == 0 );
+                    }
+                }
+                g_strfreev ( list );
+            }
+        }
+
+        if ( !show ) {
+            g_debug ( "[%s] [%s] Adding desktop file to disabled list: 'OnlyShowIn'/'NotShowIn' keys don't match current desktop", id, path );
+            g_key_file_free ( kf );
+            g_hash_table_add ( pd->disabled_entries, g_strdup ( id ) );
+            return FALSE;
+        }
+    }
     // Skip entries that have NoDisplay set.
-    if ( g_key_file_get_boolean ( kf, "Desktop Entry", "NoDisplay", NULL ) ) {
-        g_debug ( "Adding desktop file: %s to disabled list because: NoDisplay", path );
+    if ( g_key_file_get_boolean ( kf, DRUN_GROUP_NAME, "NoDisplay", NULL ) ) {
+        g_debug ( "[%s] [%s] Adding desktop file to disabled list: 'NoDisplay' key is true", id, path );
         g_key_file_free ( kf );
         g_hash_table_add ( pd->disabled_entries, g_strdup ( id ) );
         return FALSE;
     }
     // We need Exec, don't support DBusActivatable
-    if ( !g_key_file_has_key ( kf, "Desktop Entry", "Exec", NULL ) ) {
-        g_debug ( "Unsupported DesktopFile: '%s', no 'Exec' key present.", path );
+    if ( !g_key_file_has_key ( kf, DRUN_GROUP_NAME, "Exec", NULL ) ) {
+        g_debug ( "[%s] [%s] Unsupported desktop file: no 'Exec' key present.", id, path );
         g_key_file_free ( kf );
         return FALSE;
     }
 
-    if ( g_key_file_has_key ( kf, "Desktop Entry", "TryExec", NULL ) ) {
-        char *te = g_key_file_get_string ( kf, "Desktop Entry", "TryExec", NULL );
+    if ( g_key_file_has_key ( kf, DRUN_GROUP_NAME, "TryExec", NULL ) ) {
+        char *te = g_key_file_get_string ( kf, DRUN_GROUP_NAME, "TryExec", NULL );
         if ( !g_path_is_absolute ( te ) ) {
             char *fp = g_find_program_in_path ( te );
             if ( fp == NULL ) {
@@ -317,19 +388,41 @@ static gboolean read_desktop_file ( DRunModePrivateData *pd, const char *root, c
         pd->cmd_list_length_actual += 256;
         pd->entry_list              = g_realloc ( pd->entry_list, pd->cmd_list_length_actual * sizeof ( *( pd->entry_list ) ) );
     }
-    pd->entry_list[pd->cmd_list_length].icon_size = 0;
-    pd->entry_list[pd->cmd_list_length].root      = g_strdup ( root );
-    pd->entry_list[pd->cmd_list_length].path      = g_strdup ( path );
-    pd->entry_list[pd->cmd_list_length].app_id    = g_strndup ( basename, strlen ( basename ) - strlen ( ".desktop" ) );
-    gchar *n = g_key_file_get_locale_string ( kf, "Desktop Entry", "Name", NULL, NULL );
+    // Make sure order is preserved, this will break when cmd_list_length is bigger then INT_MAX.
+    // This is not likely to happen.
+    if ( G_UNLIKELY ( pd->cmd_list_length > INT_MAX ) ) {
+        // Default to smallest value.
+        pd->entry_list[pd->cmd_list_length].sort_index = INT_MIN;
+    }
+    else {
+        pd->entry_list[pd->cmd_list_length].sort_index = -pd->cmd_list_length;
+    }
+    pd->entry_list[pd->cmd_list_length].icon_size  = 0;
+    pd->entry_list[pd->cmd_list_length].root       = g_strdup ( root );
+    pd->entry_list[pd->cmd_list_length].path       = g_strdup ( path );
+    pd->entry_list[pd->cmd_list_length].desktop_id = g_strdup ( id );
+    pd->entry_list[pd->cmd_list_length].app_id     = g_strndup ( basename, strlen ( basename ) - strlen ( ".desktop" ) );
+    gchar *n = g_key_file_get_locale_string ( kf, DRUN_GROUP_NAME, "Name", NULL, NULL );
     pd->entry_list[pd->cmd_list_length].name = n;
-    gchar *gn = g_key_file_get_locale_string ( kf, "Desktop Entry", "GenericName", NULL, NULL );
+    gchar *gn = g_key_file_get_locale_string ( kf, DRUN_GROUP_NAME, "GenericName", NULL, NULL );
     pd->entry_list[pd->cmd_list_length].generic_name = gn;
-    pd->entry_list[pd->cmd_list_length].categories   = g_key_file_get_locale_string_list ( kf, "Desktop Entry", "Categories", NULL, NULL, NULL );
-    pd->entry_list[pd->cmd_list_length].exec         = g_key_file_get_string ( kf, "Desktop Entry", "Exec", NULL );
+    if ( matching_entry_fields[DRUN_MATCH_FIELD_CATEGORIES].enabled ) {
+        pd->entry_list[pd->cmd_list_length].categories = g_key_file_get_locale_string_list ( kf, DRUN_GROUP_NAME, "Categories", NULL, NULL, NULL );
+    }
+    else {
+        pd->entry_list[pd->cmd_list_length].categories = NULL;
+    }
+    pd->entry_list[pd->cmd_list_length].exec = g_key_file_get_string ( kf, DRUN_GROUP_NAME, "Exec", NULL );
 
+    if ( matching_entry_fields[DRUN_MATCH_FIELD_COMMENT].enabled ) {
+        pd->entry_list[pd->cmd_list_length].comment = g_key_file_get_locale_string ( kf,
+                                                                                     DRUN_GROUP_NAME, "Comment", NULL, NULL );
+    }
+    else {
+        pd->entry_list[pd->cmd_list_length].comment = NULL;
+    }
     if ( config.show_icons ) {
-        pd->entry_list[pd->cmd_list_length].icon_name = g_key_file_get_locale_string ( kf, "Desktop Entry", "Icon", NULL, NULL );
+        pd->entry_list[pd->cmd_list_length].icon_name = g_key_file_get_locale_string ( kf, DRUN_GROUP_NAME, "Icon", NULL, NULL );
     }
     else{
         pd->entry_list[pd->cmd_list_length].icon_name = NULL;
@@ -340,6 +433,7 @@ static gboolean read_desktop_file ( DRunModePrivateData *pd, const char *root, c
     pd->entry_list[pd->cmd_list_length].key_file = kf;
     // We don't want to parse items with this id anymore.
     g_hash_table_add ( pd->disabled_entries, g_strdup ( id ) );
+    g_debug ( "[%s] Using file %s.", id, path );
     ( pd->cmd_list_length )++;
     return TRUE;
 }
@@ -416,36 +510,46 @@ static void walk_dir ( DRunModePrivateData *pd, const char *root, const char *di
 static void delete_entry_history ( const DRunModeEntry *entry )
 {
     char *path = g_build_filename ( cache_dir, DRUN_CACHE_FILE, NULL );
-    char *key  = g_strdup_printf ( "%s:::%s", entry->root, entry->path );
-    history_remove ( path, key );
-    g_free ( key );
+    history_remove ( path, entry->desktop_id );
     g_free ( path );
 }
 
 static void get_apps_history ( DRunModePrivateData *pd )
 {
+    TICK_N ( "Start drun history" );
     unsigned int length = 0;
     gchar        *path  = g_build_filename ( cache_dir, DRUN_CACHE_FILE, NULL );
     gchar        **retv = history_get_list ( path, &length );
     for ( unsigned int index = 0; index < length; index++ ) {
-        char **st = g_strsplit ( retv[index], ":::", 2 );
-        if ( st && st[0] && st[1] ) {
-            const gchar *basename = g_utf8_strrchr ( st[1], -1, G_DIR_SEPARATOR );
-            if ( basename == NULL || !read_desktop_file ( pd, st[0], st[1], ++basename ) ) {
-                history_remove ( path, retv[index] );
+        for ( size_t i = 0; i < pd->cmd_list_length; i++ ) {
+            if ( g_strcmp0 ( pd->entry_list[i].desktop_id, retv[index] ) == 0 ) {
+                unsigned int sort_index = length - index;
+                if ( G_LIKELY ( sort_index < INT_MAX ) ) {
+                    pd->entry_list[i].sort_index = sort_index;
+                }
+                else {
+                    // This won't sort right anymore, but never gonna hit it anyway.
+                    pd->entry_list[i].sort_index = INT_MAX;
+                }
             }
         }
-        g_strfreev ( st );
     }
     g_strfreev ( retv );
     g_free ( path );
-    pd->history_length = pd->cmd_list_length;
+    TICK_N ( "Stop drun history" );
+}
+
+static gint drun_int_sort_list ( gconstpointer a, gconstpointer b, G_GNUC_UNUSED gpointer user_data )
+{
+    DRunModeEntry *da = (DRunModeEntry *) a;
+    DRunModeEntry *db = (DRunModeEntry *) b;
+
+    return db->sort_index - da->sort_index;
 }
 
 static void get_apps ( DRunModePrivateData *pd )
 {
     TICK_N ( "Get Desktop apps (start)" );
-    get_apps_history ( pd );
 
     gchar *dir;
     // First read the user directory.
@@ -471,6 +575,11 @@ static void get_apps ( DRunModePrivateData *pd )
         }
     }
     TICK_N ( "Get Desktop apps (system dirs)" );
+    get_apps_history ( pd );
+
+    g_qsort_with_data ( pd->entry_list, pd->cmd_list_length, sizeof ( DRunModeEntry ), drun_int_sort_list, NULL );
+
+    TICK_N ( "Sorting done." );
 }
 
 static void drun_icon_fetch ( gpointer data, gpointer user_data )
@@ -488,13 +597,21 @@ static void drun_icon_fetch ( gpointer data, gpointer user_data )
     if ( dr->icon_name == NULL ) {
         return;
     }
-    gchar *icon_path = nk_xdg_theme_get_icon ( pd->xdg_context, themes, NULL, dr->icon_name, dr->icon_size, 1, TRUE );
-    if ( icon_path == NULL ) {
-        g_debug ( "Failed to get Icon %s(%d): n/a", dr->icon_name, dr->icon_size  );
-        return;
+    const gchar *icon_path;
+    gchar       *icon_path_ = NULL;
+
+    if ( g_path_is_absolute ( dr->icon_name ) ) {
+        icon_path = dr->icon_name;
     }
-    else{
-        g_debug ( "Found Icon %s(%d): %s", dr->icon_name, dr->icon_size, icon_path  );
+    else {
+        icon_path = icon_path_ = nk_xdg_theme_get_icon ( pd->xdg_context, themes, NULL, dr->icon_name, dr->icon_size, 1, TRUE );
+        if ( icon_path_ == NULL ) {
+            g_debug ( "Failed to get Icon %s(%d): n/a", dr->icon_name, dr->icon_size  );
+            return;
+        }
+        else{
+            g_debug ( "Found Icon %s(%d): %s", dr->icon_name, dr->icon_size, icon_path  );
+        }
     }
     cairo_surface_t *icon_surf = NULL;
     if ( g_str_has_suffix ( icon_path, ".png" ) ) {
@@ -515,29 +632,73 @@ static void drun_icon_fetch ( gpointer data, gpointer user_data )
         }
         dr->icon = icon_surf;
     }
-    g_free ( icon_path );
+    g_free ( icon_path_ );
     rofi_view_reload ();
 }
 
+static void drun_mode_parse_entry_fields ()
+{
+    char               *savept = NULL;
+    // Make a copy, as strtok will modify it.
+    char               *switcher_str = g_strdup ( config.drun_match_fields );
+    const char * const sep           = ",#";
+    // Split token on ','. This modifies switcher_str.
+    for ( unsigned int i = 0; i < DRUN_MATCH_NUM_FIELDS; i++ ) {
+        matching_entry_fields[i].enabled = FALSE;
+    }
+    for ( char *token = strtok_r ( switcher_str, sep, &savept ); token != NULL;
+          token = strtok_r ( NULL, sep, &savept ) ) {
+        if ( strcmp ( token, "all" ) == 0 ) {
+            for ( unsigned int i = 0; i < DRUN_MATCH_NUM_FIELDS; i++ ) {
+                matching_entry_fields[i].enabled = TRUE;
+            }
+            break;
+        }
+        else {
+            gboolean matched = FALSE;
+            for ( unsigned int i = 0; i < DRUN_MATCH_NUM_FIELDS; i++ ) {
+                const char * entry_name = matching_entry_fields[i].entry_field_name;
+                if ( g_ascii_strcasecmp ( token, entry_name ) == 0 ) {
+                    matching_entry_fields[i].enabled = TRUE;
+                    matched                          = TRUE;
+                }
+            }
+            if ( !matched ) {
+                g_warning ( "Invalid entry name :%s", token );
+            }
+        }
+    }
+    // Free string that was modified by strtok_r
+    g_free ( switcher_str );
+}
+
 static int drun_mode_init ( Mode *sw )
 {
-    if ( mode_get_private_data ( sw ) == NULL ) {
-        static const gchar * const drun_icon_fallback_themes[] = {
-            "Adwaita",
-            "gnome",
-            NULL
-        };
-        const gchar                *themes[2] = {
-            config.drun_icon_theme,
-            NULL
-        };
-        DRunModePrivateData        *pd = g_malloc0 ( sizeof ( *pd ) );
-        pd->disabled_entries = g_hash_table_new_full ( g_str_hash, g_str_equal, g_free, NULL );
-        mode_set_private_data ( sw, (void *) pd );
-        pd->xdg_context = nk_xdg_theme_context_new ( drun_icon_fallback_themes, NULL );
-        nk_xdg_theme_preload_themes_icon ( pd->xdg_context, themes );
-        get_apps ( pd );
+    if ( mode_get_private_data ( sw ) != NULL ) {
+        return TRUE;
     }
+
+    static const gchar * const drun_icon_fallback_themes[] = {
+        "Adwaita",
+        "gnome",
+        NULL
+    };
+    const gchar                *themes[2] = {
+        config.drun_icon_theme,
+        NULL
+    };
+    DRunModePrivateData        *pd = g_malloc0 ( sizeof ( *pd ) );
+    pd->disabled_entries = g_hash_table_new_full ( g_str_hash, g_str_equal, g_free, NULL );
+    mode_set_private_data ( sw, (void *) pd );
+    // current destkop
+    const char *current_desktop = g_getenv ( "XDG_CURRENT_DESKTOP" );
+    pd->current_desktop_list = current_desktop ? g_strsplit ( current_desktop, ":", 0 ) : NULL;
+
+    // Theme
+    pd->xdg_context = nk_xdg_theme_context_new ( drun_icon_fallback_themes, NULL );
+    nk_xdg_theme_preload_themes_icon ( pd->xdg_context, themes );
+    drun_mode_parse_entry_fields ();
+    get_apps ( pd );
     return TRUE;
 }
 static void drun_entry_clear ( DRunModeEntry *e )
@@ -545,6 +706,7 @@ static void drun_entry_clear ( DRunModeEntry *e )
     g_free ( e->root );
     g_free ( e->path );
     g_free ( e->app_id );
+    g_free ( e->desktop_id );
     if ( e->icon != NULL ) {
         cairo_surface_destroy ( e->icon );
     }
@@ -552,6 +714,7 @@ static void drun_entry_clear ( DRunModeEntry *e )
     g_free ( e->exec );
     g_free ( e->name );
     g_free ( e->generic_name );
+    g_free ( e->comment );
     g_strfreev ( e->categories );
     g_key_file_free ( e->key_file );
 }
@@ -581,7 +744,8 @@ static ModeMode drun_mode_result ( Mode *sw, int mretv, char **input, unsigned i
         helper_execute_command ( NULL, *input, run_in_term, run_in_term ? &context : NULL );
     }
     else if ( ( mretv & MENU_ENTRY_DELETE ) && selected_line < rmpd->cmd_list_length ) {
-        if ( selected_line < rmpd->history_length ) {
+        // Possitive sort index means it is in history.
+        if ( rmpd->entry_list[selected_line].sort_index >= 0 ) {
             if ( rmpd->pool ) {
                 g_thread_pool_free ( rmpd->pool, TRUE, TRUE );
                 rmpd->pool = NULL;
@@ -610,6 +774,8 @@ static void drun_mode_destroy ( Mode *sw )
         g_hash_table_destroy ( rmpd->disabled_entries );
         g_free ( rmpd->entry_list );
         nk_xdg_theme_context_free ( rmpd->xdg_context );
+
+        g_strfreev ( rmpd->current_desktop_list );
         g_free ( rmpd );
         mode_set_private_data ( sw, NULL );
     }
@@ -667,33 +833,45 @@ static char *drun_get_completion ( const Mode *sw, unsigned int index )
     }
 }
 
-static int drun_token_match ( const Mode *data, GRegex **tokens, unsigned int index )
+static int drun_token_match ( const Mode *data, rofi_int_matcher **tokens, unsigned int index )
 {
     DRunModePrivateData *rmpd = (DRunModePrivateData *) mode_get_private_data ( data );
     int                 match = 1;
     if ( tokens ) {
         for ( int j = 0; match && tokens != NULL && tokens[j] != NULL; j++ ) {
-            int    test        = 0;
-            GRegex *ftokens[2] = { tokens[j], NULL };
+            int              test        = 0;
+            rofi_int_matcher *ftokens[2] = { tokens[j], NULL };
             // Match name
-            if ( rmpd->entry_list[index].name &&
-                 helper_token_match ( ftokens, rmpd->entry_list[index].name ) ) {
-                test = 1;
+            if ( matching_entry_fields[DRUN_MATCH_FIELD_NAME].enabled ) {
+                if ( rmpd->entry_list[index].name ) {
+                    test = helper_token_match ( ftokens, rmpd->entry_list[index].name );
+                }
             }
-            // Match generic name
-            if ( !test && rmpd->entry_list[index].generic_name &&
-                 helper_token_match ( ftokens, rmpd->entry_list[index].generic_name ) ) {
-                test = 1;
+            if ( matching_entry_fields[DRUN_MATCH_FIELD_GENERIC].enabled ) {
+                // Match generic name
+                if ( test == tokens[j]->invert && rmpd->entry_list[index].generic_name ) {
+                    test = helper_token_match ( ftokens, rmpd->entry_list[index].generic_name );
+                }
             }
-            // Match executable name.
-            if ( !test && helper_token_match ( ftokens, rmpd->entry_list[index].exec ) ) {
-                test = 1;
+            if ( matching_entry_fields[DRUN_MATCH_FIELD_EXEC].enabled ) {
+                // Match executable name.
+                if ( test == tokens[j]->invert  ) {
+                    test = helper_token_match ( ftokens, rmpd->entry_list[index].exec );
+                }
             }
-            // Match against category.
-            if ( !test ) {
-                gchar **list = rmpd->entry_list[index].categories;
-                for ( int iter = 0; !test && list && list[iter]; iter++ ) {
-                    test = helper_token_match ( ftokens, list[iter] );
+            if ( matching_entry_fields[DRUN_MATCH_FIELD_CATEGORIES].enabled ) {
+                // Match against category.
+                if ( test == tokens[j]->invert ) {
+                    gchar **list = rmpd->entry_list[index].categories;
+                    for ( int iter = 0; test == tokens[j]->invert && list && list[iter]; iter++ ) {
+                        test = helper_token_match ( ftokens, list[iter] );
+                    }
+                }
+            }
+            if ( matching_entry_fields[DRUN_MATCH_FIELD_COMMENT].enabled ) {
+                // Match executable name.
+                if ( test == tokens[j]->invert && rmpd->entry_list[index].comment ) {
+                    test = helper_token_match ( ftokens, rmpd->entry_list[index].comment );
                 }
             }
             if ( test == 0 ) {
@@ -701,6 +879,7 @@ static int drun_token_match ( const Mode *data, GRegex **tokens, unsigned int in
             }
         }
     }
+
     return match;
 }
 
