@@ -152,10 +152,11 @@ int helper_parse_setup ( char * string, char ***output, int *length, ... )
     return FALSE;
 }
 
-void tokenize_free ( GRegex ** tokens )
+void helper_tokenize_free ( rofi_int_matcher ** tokens )
 {
     for ( size_t i = 0; tokens && tokens[i]; i++ ) {
-        g_regex_unref ( (GRegex *) tokens[i] );
+        g_regex_unref ( (GRegex *) tokens[i]->regex );
+        g_free ( tokens[i] );
     }
     g_free ( tokens );
 }
@@ -214,10 +215,15 @@ static inline GRegex * R ( const char *s, int case_sensitive  )
     return g_regex_new ( s, G_REGEX_OPTIMIZE | ( ( case_sensitive ) ? 0 : G_REGEX_CASELESS ), 0, NULL );
 }
 
-static GRegex * create_regex ( const char *input, int case_sensitive )
+static rofi_int_matcher * create_regex ( const char *input, int case_sensitive )
 {
-    GRegex * retv = NULL;
-    gchar  *r;
+    GRegex           * retv = NULL;
+    gchar            *r;
+    rofi_int_matcher *rv = g_malloc0 ( sizeof ( rofi_int_matcher ) );
+    if ( input && input[0] == '-' ) {
+        rv->invert = 1;
+        input++;
+    }
     switch ( config.matching_method )
     {
     case MM_GLOB:
@@ -244,9 +250,10 @@ static GRegex * create_regex ( const char *input, int case_sensitive )
         g_free ( r );
         break;
     }
-    return retv;
+    rv->regex = retv;
+    return rv;
 }
-GRegex **tokenize ( const char *input, int case_sensitive )
+rofi_int_matcher **helper_tokenize ( const char *input, int case_sensitive )
 {
     if ( input == NULL ) {
         return NULL;
@@ -256,11 +263,11 @@ GRegex **tokenize ( const char *input, int case_sensitive )
         return NULL;
     }
 
-    char   *saveptr = NULL, *token;
-    GRegex **retv = NULL;
+    char             *saveptr = NULL, *token;
+    rofi_int_matcher **retv = NULL;
     if ( !config.tokenize ) {
-        retv    = g_malloc0 ( sizeof ( GRegex* ) * 2 );
-        retv[0] = (GRegex *) create_regex ( input, case_sensitive );
+        retv    = g_malloc0 ( sizeof ( rofi_int_matcher* ) * 2 );
+        retv[0] = create_regex ( input, case_sensitive );
         return retv;
     }
 
@@ -274,8 +281,8 @@ GRegex **tokenize ( const char *input, int case_sensitive )
     // strtok should still be valid for utf8.
     const char * const sep = " ";
     for ( token = strtok_r ( str, sep, &saveptr ); token != NULL; token = strtok_r ( NULL, sep, &saveptr ) ) {
-        retv                 = g_realloc ( retv, sizeof ( GRegex* ) * ( num_tokens + 2 ) );
-        retv[num_tokens]     = (GRegex *) create_regex ( token, case_sensitive );
+        retv                 = g_realloc ( retv, sizeof ( rofi_int_matcher* ) * ( num_tokens + 2 ) );
+        retv[num_tokens]     = create_regex ( token, case_sensitive );
         retv[num_tokens + 1] = NULL;
         num_tokens++;
     }
@@ -400,13 +407,16 @@ int find_arg_char ( const char * const key, char *val )
     return FALSE;
 }
 
-PangoAttrList *helper_token_match_get_pango_attr ( RofiHighlightColorStyle th, GRegex **tokens, const char *input, PangoAttrList *retv )
+PangoAttrList *helper_token_match_get_pango_attr ( RofiHighlightColorStyle th, rofi_int_matcher**tokens, const char *input, PangoAttrList *retv )
 {
     // Do a tokenized match.
     if ( tokens ) {
         for ( int j = 0; tokens[j]; j++ ) {
             GMatchInfo *gmi = NULL;
-            g_regex_match ( (GRegex *) tokens[j], input, G_REGEX_MATCH_PARTIAL, &gmi );
+            if ( tokens[j]->invert ) {
+                continue;
+            }
+            g_regex_match ( tokens[j]->regex, input, G_REGEX_MATCH_PARTIAL, &gmi );
             while ( g_match_info_matches ( gmi ) ) {
                 int count = g_match_info_get_match_count ( gmi );
                 for ( int index = ( count > 1 ) ? 1 : 0; index < count; index++ ) {
@@ -460,13 +470,14 @@ PangoAttrList *helper_token_match_get_pango_attr ( RofiHighlightColorStyle th, G
     return retv;
 }
 
-int helper_token_match ( GRegex * const *tokens, const char *input )
+int helper_token_match ( rofi_int_matcher* const *tokens, const char *input )
 {
     int match = TRUE;
     // Do a tokenized match.
     if ( tokens ) {
         for ( int j = 0; match && tokens[j]; j++ ) {
-            match = g_regex_match ( (const GRegex *) tokens[j], input, 0, NULL );
+            match  = g_regex_match ( tokens[j]->regex, input, 0, NULL );
+            match ^= tokens[j]->invert;
         }
     }
     return match;
@@ -854,7 +865,7 @@ static int rofi_scorer_get_score_for ( enum CharClass prev, enum CharClass curr
  * @param pattern   The user input to match against.
  * @param plen      Pattern length.
  * @param str       The input to match against pattern.
- * @param slen      Lenght of str.
+ * @param slen      Length of str.
  *
  *  rofi_scorer_fuzzy_evaluate implements a global sequence alignment algorithm to find the maximum accumulated score by
  *  aligning `pattern` to `str`. It applies when `pattern` is a subsequence of `str`.
@@ -1114,3 +1125,93 @@ cairo_surface_t* cairo_image_surface_create_from_svg ( const gchar* file, int he
 
     return surface;
 }
+
+static void parse_pair ( char  *input, rofi_range_pair  *item )
+{
+    int                index = 0;
+    const char * const sep   = "-";
+    for ( char *token = strsep ( &input, sep ); token != NULL; token = strsep ( &input, sep ) ) {
+        if ( index == 0 ) {
+            item->start = item->stop = (unsigned int) strtoul ( token, NULL, 10 );
+            index++;
+        }
+        else {
+            if ( token[0] == '\0' ) {
+                item->stop = 0xFFFFFFFF;
+            }
+            else{
+                item->stop = (unsigned int) strtoul ( token, NULL, 10 );
+            }
+        }
+    }
+}
+void parse_ranges ( char *input, rofi_range_pair **list, unsigned int *length )
+{
+    char *endp;
+    if ( input == NULL ) {
+        return;
+    }
+    const char *const sep = ",";
+    for ( char *token = strtok_r ( input, sep, &endp ); token != NULL; token = strtok_r ( NULL, sep, &endp ) ) {
+        // Make space.
+        *list = g_realloc ( ( *list ), ( ( *length ) + 1 ) * sizeof ( struct rofi_range_pair ) );
+        // Parse a single pair.
+        parse_pair ( token, &( ( *list )[*length] ) );
+
+        ( *length )++;
+    }
+}
+/**
+ * @param format The format string used. See below for possible syntax.
+ * @param string The selected entry.
+ * @param selected_line The selected line index.
+ * @param filter The entered filter.
+ *
+ * Function that outputs the selected line in the user-specified format.
+ * Currently the following formats are supported:
+ *   * i: Print the index (0-(N-1))
+ *   * d: Print the index (1-N)
+ *   * s: Print input string.
+ *   * q: Print quoted input string.
+ *   * f: Print the entered filter.
+ *   * F: Print the entered filter, quoted
+ *
+ * This functions outputs the formatted string to stdout, appends a newline (\n) character and
+ * calls flush on the file descriptor.
+ */
+void rofi_output_formatted_line ( const char *format, const char *string, int selected_line, const char *filter )
+{
+    for ( int i = 0; format && format[i]; i++ ) {
+        if ( format[i] == 'i' ) {
+            fprintf ( stdout, "%d", selected_line );
+        }
+        else if ( format[i] == 'd' ) {
+            fprintf ( stdout, "%d", ( selected_line + 1 ) );
+        }
+        else if ( format[i] == 's' ) {
+            fputs ( string, stdout );
+        }
+        else if ( format[i] == 'q' ) {
+            char *quote = g_shell_quote ( string );
+            fputs ( quote, stdout );
+            g_free ( quote );
+        }
+        else if ( format[i] == 'f' ) {
+            if ( filter ) {
+                fputs ( filter, stdout );
+            }
+        }
+        else if ( format[i] == 'F' ) {
+            if ( filter ) {
+                char *quote = g_shell_quote ( filter );
+                fputs ( quote, stdout );
+                g_free ( quote );
+            }
+        }
+        else {
+            fputc ( format[i], stdout );
+        }
+    }
+    fputc ( '\n', stdout );
+    fflush ( stdout );
+}
