@@ -38,8 +38,6 @@
 #include "history.h"
 #include "settings.h"
 
-#define HISTORY_MAX_ENTRIES    25
-
 /**
  * History element
  */
@@ -70,7 +68,7 @@ static void __history_write_element_list ( FILE *fd, _element **list, unsigned i
     int min_value = list[length - 1]->index;
 
     // Set the max length of the list.
-    length = ( length > HISTORY_MAX_ENTRIES ) ? HISTORY_MAX_ENTRIES : length;
+    length = ( length > config.max_history_size ) ? config.max_history_size : length;
 
     // Write out entries.
     for ( unsigned int iter = 0; iter < length; iter++ ) {
@@ -78,9 +76,50 @@ static void __history_write_element_list ( FILE *fd, _element **list, unsigned i
     }
 }
 
+static char ** __history_get_element_list_fields ( FILE *fd, unsigned int *length )
+{
+    unsigned int real_length = 0;
+    char         **retv      = NULL;;
+    if  ( length == NULL ) {
+        return NULL;
+    }
+    *length = 0;
+
+    if ( fd == NULL ) {
+        return NULL;
+    }
+    char    *buffer       = NULL;
+    size_t  buffer_length = 0;
+    ssize_t l             = 0;
+    while ( ( l = getline ( &buffer, &buffer_length, fd ) ) > 0 ) {
+        // Jump to the first space.
+        const char *start = strchr ( buffer, ' ' );
+        // not found, skip.
+        if ( start == NULL ) {
+            continue;
+        }
+        start++;
+        // remove trailing \n
+        buffer[l - 1] = '\0';
+        if ( real_length < ( *length + 2 ) ) {
+            real_length += 15;
+            // Resize and check.
+            retv = g_realloc ( retv, ( real_length ) * sizeof ( char* ) );
+        }
+        // Parse the number of times.
+        retv[( *length )] = g_strndup ( start, l - 1 - ( start - buffer ) );
+        // Force trailing '\0'
+        retv[( *length ) + 1] = NULL;
+
+        ( *length )++;
+    }
+    return retv;
+}
+
 static _element ** __history_get_element_list ( FILE *fd, unsigned int *length )
 {
-    _element **retv = NULL;
+    unsigned int real_length = 0;
+    _element     **retv      = NULL;
 
     if  ( length == NULL ) {
         return NULL;
@@ -108,8 +147,11 @@ static _element ** __history_get_element_list ( FILE *fd, unsigned int *length )
         if ( ( l - ( start - buffer ) ) < 2 ) {
             continue;
         }
-        // Resize and check.
-        retv = g_realloc ( retv, ( *length + 2 ) * sizeof ( _element* ) );
+        if ( real_length < ( *length + 2 ) ) {
+            real_length += 15;
+            // Resize and check.
+            retv = g_realloc ( retv, ( real_length ) * sizeof ( _element* ) );
+        }
 
         retv[( *length )] = g_malloc ( sizeof ( _element ) );
 
@@ -270,10 +312,9 @@ char ** history_get_list ( const char *filename, unsigned int *length )
     if ( config.disable_history ) {
         return NULL;
     }
-    _element **list = NULL;
-    char     **retv = NULL;
+    char **retv = NULL;
     // Open file.
-    FILE     *fd = g_fopen ( filename, "r" );
+    FILE *fd = g_fopen ( filename, "r" );
     if ( fd == NULL ) {
         // File that does not exists is not an error, so ignore it.
         // Everything else? panic.
@@ -283,19 +324,7 @@ char ** history_get_list ( const char *filename, unsigned int *length )
         return NULL;
     }
     // Get list.
-    list = __history_get_element_list ( fd, length );
-
-    // Copy list in right format.
-    // Lists are always short, so performance should not be an issue.
-    if ( ( *length ) > 0 ) {
-        retv = g_malloc ( ( ( *length ) + 1 ) * sizeof ( char * ) );
-        for ( unsigned int iter = 0; iter < ( *length ); iter++ ) {
-            retv[iter] = ( list[iter]->name );
-            g_free ( list[iter] );
-        }
-        retv[( *length )] = NULL;
-        g_free ( list );
-    }
+    retv = __history_get_element_list_fields ( fd, length );
 
     // Close file, if fails let user know on stderr.
     if ( fclose ( fd ) != 0 ) {
