@@ -244,14 +244,7 @@ static void rofi_view_update_prompt ( RofiViewState *state )
 {
     if ( state->prompt ) {
         const char *str = mode_get_display_name ( state->sw );
-        if ( ( state->menu_flags & MENU_PROMPT_COLON ) != 0 ) {
-            char *pr = g_strconcat ( str, ":", NULL );
-            textbox_text ( state->prompt, pr );
-            g_free ( pr );
-        }
-        else {
-            textbox_text ( state->prompt, str );
-        }
+        textbox_text ( state->prompt, str );
     }
 }
 
@@ -308,31 +301,31 @@ static void rofi_view_calculate_window_position ( RofiViewState *state )
     {
     case WL_NORTH_WEST:
         state->x = CacheState.mon.x;
-        /* FALLTHRU */
+    /* FALLTHRU */
     case WL_NORTH:
         state->y = CacheState.mon.y;
         break;
     case WL_NORTH_EAST:
         state->y = CacheState.mon.y;
-        /* FALLTHRU */
+    /* FALLTHRU */
     case WL_EAST:
         state->x = CacheState.mon.x + CacheState.mon.w;
         break;
     case WL_SOUTH_EAST:
         state->x = CacheState.mon.x + CacheState.mon.w;
-        /* FALLTHRU */
+    /* FALLTHRU */
     case WL_SOUTH:
         state->y = CacheState.mon.y + CacheState.mon.h;
         break;
     case WL_SOUTH_WEST:
         state->y = CacheState.mon.y + CacheState.mon.h;
-        /* FALLTHRU */
+    /* FALLTHRU */
     case WL_WEST:
         state->x = CacheState.mon.x;
         break;
     case WL_CENTER:
         ;
-        /* FALLTHRU */
+    /* FALLTHRU */
     default:
         break;
     }
@@ -497,7 +490,7 @@ void rofi_view_set_selected_line ( RofiViewState *state, unsigned int selected_l
 void rofi_view_free ( RofiViewState *state )
 {
     if ( state->tokens ) {
-        tokenize_free ( state->tokens );
+        helper_tokenize_free ( state->tokens );
         state->tokens = NULL;
     }
     // Do this here?
@@ -789,6 +782,8 @@ void __create_window ( MenuFlags menu_flags )
     const char *transparency = rofi_theme_get_string ( WIDGET ( win ), "transparency", NULL );
     if ( transparency ) {
         rofi_view_setup_fake_transparency ( transparency  );
+    } else if ( config.fake_transparency && config.fake_background ) {
+        rofi_view_setup_fake_transparency ( config.fake_background );
     }
     if ( xcb->sncontext != NULL ) {
         sn_launchee_context_setup_window ( xcb->sncontext, CacheState.main_window );
@@ -1019,14 +1014,14 @@ static void rofi_view_refilter ( RofiViewState *state )
         state->reload = FALSE;
     }
     if ( state->tokens ) {
-        tokenize_free ( state->tokens );
+        helper_tokenize_free ( state->tokens );
         state->tokens = NULL;
     }
     if ( state->text && strlen ( state->text->text ) > 0 ) {
         unsigned int j        = 0;
         gchar        *pattern = mode_preprocess_input ( state->sw, state->text->text );
         glong        plen     = pattern ? g_utf8_strlen ( pattern, -1 ) : 0;
-        state->tokens = tokenize ( pattern, config.case_sensitive );
+        state->tokens = helper_tokenize ( pattern, config.case_sensitive );
         /**
          * On long lists it can be beneficial to parallelize.
          * If number of threads is 1, no thread is spawn.
@@ -1364,10 +1359,10 @@ gboolean rofi_view_trigger_action ( RofiViewState *state, BindingsScope scope, g
             return FALSE;
         case WIDGET_TRIGGER_ACTION_RESULT_GRAB_MOTION_END:
             target = NULL;
-            /* FALLTHRU */
+        /* FALLTHRU */
         case WIDGET_TRIGGER_ACTION_RESULT_GRAB_MOTION_BEGIN:
             state->mouse.motion_target = target;
-            /* FALLTHRU */
+        /* FALLTHRU */
         case WIDGET_TRIGGER_ACTION_RESULT_HANDLED:
             return TRUE;
         }
@@ -1464,15 +1459,19 @@ void rofi_view_frame_callback ( void )
 
 static int rofi_view_calculate_height ( RofiViewState *state )
 {
-    unsigned int height = 0;
     if ( CacheState.fullscreen == TRUE ) {
-        height = CacheState.mon.h;
-        return height;
+        return CacheState.mon.h;
     }
 
+    RofiDistance h      = rofi_theme_get_distance ( WIDGET ( state->main_window ), "height", 0 );
+    unsigned int height = distance_get_pixel ( h, ROFI_ORIENTATION_VERTICAL );
+    // If height is set, return it.
+    if ( height > 0 ) {
+        return height;
+    }
+    // Autosize based on widgets.
     widget *main_window = WIDGET ( state->main_window );
-    height = widget_get_desired_height ( main_window );
-    return height;
+    return widget_get_desired_height ( main_window );
 }
 
 static WidgetTriggerActionResult textbox_sidebar_modi_trigger_action ( widget *wid, MouseBindingMouseDefaultAction action, G_GNUC_UNUSED gint x, G_GNUC_UNUSED gint y, G_GNUC_UNUSED void *user_data )
@@ -1621,7 +1620,7 @@ static void rofi_view_add_widget ( RofiViewState *state, widget *parent_widget,
             return;
         }
         if ( config.sidebar_mode ) {
-            state->sidebar_bar = box_create ( parent_widget, name,ROFI_ORIENTATION_HORIZONTAL );
+            state->sidebar_bar = box_create ( parent_widget, name, ROFI_ORIENTATION_HORIZONTAL );
             box_add ( (box *) parent_widget, WIDGET ( state->sidebar_bar ), FALSE );
             state->num_modi = rofi_get_num_enabled_modi ();
             state->modi     = g_malloc0 ( state->num_modi * sizeof ( textbox * ) );
@@ -1635,7 +1634,7 @@ static void rofi_view_add_widget ( RofiViewState *state, widget *parent_widget,
         }
     }
     else if (  g_ascii_strncasecmp ( name, "textbox", 7 ) == 0 ) {
-        textbox *t = textbox_create ( parent_widget, WIDGET_TYPE_TEXTBOX_TEXT, name, TB_WRAP, NORMAL, "", 0, 0 );
+        textbox *t = textbox_create ( parent_widget, WIDGET_TYPE_TEXTBOX_TEXT, name, TB_AUTOHEIGHT | TB_WRAP, NORMAL, "", 0, 0 );
         box_add ( (box *) parent_widget, WIDGET ( t ), TRUE );
     }
     else {
@@ -1692,7 +1691,7 @@ RofiViewState *rofi_view_create ( Mode *sw,
         textbox_cursor_end ( state->text );
     }
 
-    state->overlay                = textbox_create ( WIDGET ( state->main_window), WIDGET_TYPE_TEXTBOX_TEXT, "overlay", TB_AUTOWIDTH | TB_AUTOHEIGHT, URGENT, "blaat", 0.5, 0 );
+    state->overlay = textbox_create ( WIDGET ( state->main_window ), WIDGET_TYPE_TEXTBOX_TEXT, "overlay", TB_AUTOWIDTH | TB_AUTOHEIGHT, URGENT, "blaat", 0.5, 0 );
     widget_disable ( WIDGET ( state->overlay ) );
 
     // filtered list
@@ -1705,10 +1704,12 @@ RofiViewState *rofi_view_create ( Mode *sw,
     // Only needed when window is fixed size.
     if ( ( CacheState.flags & MENU_NORMAL_WINDOW ) == MENU_NORMAL_WINDOW ) {
         listview_set_fixed_num_lines ( state->list_view );
-        rofi_view_window_update_size ( state );
     }
+
+    state->height = rofi_view_calculate_height ( state );
     // Move the window to the correct x,y position.
     rofi_view_calculate_window_position ( state );
+    rofi_view_window_update_size ( state );
 
     state->quit = FALSE;
     rofi_view_refilter ( state );
@@ -1830,7 +1831,7 @@ void rofi_view_workers_initialize ( void )
         // We are allowed to have
         g_thread_pool_set_max_threads ( tpool, config.threads, &error );
     }
-    // If error occured during setup of pool, tell user and exit.
+    // If error occurred during setup of pool, tell user and exit.
     if ( error != NULL ) {
         g_warning ( "Failed to setup thread pool: '%s'", error->message );
         g_error_free ( error );
