@@ -481,6 +481,56 @@ static int monitor_active_from_winid ( xcb_drawable_t id, workarea *mon )
     }
     return FALSE;
 }
+static int monitor_active_from_id_focused ( int mon_id, workarea *mon )
+{
+    int                       retv = FALSE;
+    xcb_window_t              active_window;
+    xcb_get_property_cookie_t awc;
+    awc = xcb_ewmh_get_active_window ( &xcb->ewmh, xcb->screen_nbr );
+    if ( !xcb_ewmh_get_active_window_reply ( &xcb->ewmh, awc, &active_window, NULL ) ) {
+        g_debug ( "Failed to get active window, falling back to mouse location (-5)." );
+        return retv;
+    }
+    xcb_query_tree_cookie_t tree_cookie = xcb_query_tree ( xcb->connection, active_window );
+    xcb_query_tree_reply_t  *tree_reply = xcb_query_tree_reply ( xcb->connection, tree_cookie, NULL );
+    if ( !tree_reply ) {
+        g_debug ( "Failed to get parent window, falling back to mouse location (-5)." );
+        return retv;
+    }
+    // get geometry.
+    xcb_get_geometry_cookie_t c  = xcb_get_geometry ( xcb->connection, active_window );
+    xcb_get_geometry_reply_t  *r = xcb_get_geometry_reply ( xcb->connection, c, NULL );
+    if ( !r ) {
+        g_debug ( "Failed to get geometry of active window, falling back to mouse location (-5)." );
+        free ( tree_reply );
+        return retv;
+    }
+    xcb_translate_coordinates_cookie_t ct = xcb_translate_coordinates ( xcb->connection, tree_reply->parent, r->root, r->x, r->y );
+    xcb_translate_coordinates_reply_t  *t = xcb_translate_coordinates_reply ( xcb->connection, ct, NULL );
+    if ( t ) {
+        if ( mon_id == -2 ) {
+            // place the menu above the window
+            // if some window is focused, place menu above window, else fall
+            // back to selected monitor.
+            mon->x = t->dst_x - r->x;
+            mon->y = t->dst_y - r->y;
+            mon->w = r->width;
+            mon->h = r->height;
+            retv   = TRUE;
+        }
+        else if ( mon_id == -4 ) {
+            monitor_dimensions ( t->dst_x, t->dst_y, mon );
+            retv = TRUE;
+        }
+        free ( t );
+    }
+    else {
+        g_debug ( "Failed to get translate position of active window, falling back to mouse location (-5)." );
+    }
+    free ( r );
+    free ( tree_reply );
+    return retv;
+}
 static int monitor_active_from_id ( int mon_id, workarea *mon )
 {
     xcb_window_t root = xcb->screen->root;
@@ -524,47 +574,8 @@ static int monitor_active_from_id ( int mon_id, workarea *mon )
         }
     }
     else if ( mon_id == -2 || mon_id == -4 ) {
-        xcb_window_t              active_window;
-        xcb_get_property_cookie_t awc;
-        awc = xcb_ewmh_get_active_window ( &xcb->ewmh, xcb->screen_nbr );
-        if ( xcb_ewmh_get_active_window_reply ( &xcb->ewmh, awc, &active_window, NULL ) ) {
-            // get geometry.
-            xcb_get_geometry_cookie_t c  = xcb_get_geometry ( xcb->connection, active_window );
-            xcb_get_geometry_reply_t  *r = xcb_get_geometry_reply ( xcb->connection, c, NULL );
-            if ( r ) {
-                xcb_translate_coordinates_cookie_t ct = xcb_translate_coordinates ( xcb->connection, active_window, root, r->x, r->y );
-                xcb_translate_coordinates_reply_t  *t = xcb_translate_coordinates_reply ( xcb->connection, ct, NULL );
-                if ( t ) {
-                    if ( mon_id == -2 ) {
-                        // place the menu above the window
-                        // if some window is focused, place menu above window, else fall
-                        // back to selected monitor.
-                        mon->x = t->dst_x - r->x;
-                        mon->y = t->dst_y - r->y;
-                        mon->w = r->width;
-                        mon->h = r->height;
-                        free ( r );
-                        free ( t );
-                        return TRUE;
-                    }
-                    else if ( mon_id == -4 ) {
-                        monitor_dimensions ( t->dst_x, t->dst_y, mon );
-                        free ( r );
-                        free ( t );
-                        return TRUE;
-                    }
-                }
-                else {
-                    g_debug ( "Failed to get translate position of active window, falling back to mouse location (-5)." );
-                }
-                free ( r );
-            }
-            else {
-                g_debug ( "Failed to get geometry of active window, falling back to mouse location (-5)." );
-            }
-        }
-        else {
-            g_debug ( "Failed to get active window, falling back to mouse location (-5)." );
+        if ( monitor_active_from_id_focused ( mon_id, mon ) ) {
+            return TRUE;
         }
     }
     // Monitor that has mouse pointer.
@@ -660,6 +671,59 @@ static void rofi_view_paste ( RofiViewState *state, xcb_selection_notify_event_t
     }
 }
 
+static gboolean x11_button_to_nk_bindings_button ( guint32 x11_button, NkBindingsMouseButton *button )
+{
+    switch ( x11_button )
+    {
+    case 1:
+        *button = NK_BINDINGS_MOUSE_BUTTON_PRIMARY;
+        break;
+    case 3:
+        *button = NK_BINDINGS_MOUSE_BUTTON_SECONDARY;
+        break;
+    case 2:
+        *button = NK_BINDINGS_MOUSE_BUTTON_MIDDLE;
+        break;
+    case 8:
+        *button = NK_BINDINGS_MOUSE_BUTTON_BACK;
+        break;
+    case 9:
+        *button = NK_BINDINGS_MOUSE_BUTTON_FORWARD;
+        break;
+    case 4:
+    case 5:
+    case 6:
+    case 7:
+        return FALSE;
+    default:
+        *button = NK_BINDINGS_MOUSE_BUTTON_EXTRA + x11_button;
+    }
+    return TRUE;
+}
+
+static gboolean x11_button_to_nk_bindings_scroll ( guint32 x11_button, NkBindingsScrollAxis *axis, gint32 *steps )
+{
+    *steps = 1;
+    switch ( x11_button )
+    {
+    case 4:
+        *steps = -1;
+        /* fallthrough */
+    case 5:
+        *axis = NK_BINDINGS_SCROLL_AXIS_VERTICAL;
+        break;
+    case 6:
+        *steps = -1;
+        /* fallthrough */
+    case 7:
+        *axis = NK_BINDINGS_SCROLL_AXIS_HORIZONTAL;
+        break;
+    default:
+        return FALSE;
+    }
+    return TRUE;
+}
+
 /**
  * Process X11 events in the main-loop (gui-thread) of the application.
  */
@@ -693,16 +757,26 @@ static void main_loop_x11_event_handler_view ( xcb_generic_event_t *event )
     case XCB_BUTTON_PRESS:
     {
         xcb_button_press_event_t *bpe = (xcb_button_press_event_t *) event;
+        NkBindingsMouseButton button;
+        NkBindingsScrollAxis axis;
+        gint32 steps;
+
         xcb->last_timestamp = bpe->time;
         rofi_view_handle_mouse_motion ( state, bpe->event_x, bpe->event_y );
-        nk_bindings_seat_handle_button ( xcb->bindings_seat, bpe->detail, NK_BINDINGS_BUTTON_STATE_PRESS, bpe->time );
+        if ( x11_button_to_nk_bindings_button ( bpe->detail, &button ) )
+            nk_bindings_seat_handle_button ( xcb->bindings_seat, NULL, button, NK_BINDINGS_BUTTON_STATE_PRESS, bpe->time );
+        else if ( x11_button_to_nk_bindings_scroll ( bpe->detail, &axis, &steps) )
+            nk_bindings_seat_handle_scroll ( xcb->bindings_seat, NULL, axis, steps );
         break;
     }
     case XCB_BUTTON_RELEASE:
     {
         xcb_button_release_event_t *bre = (xcb_button_release_event_t *) event;
+        NkBindingsMouseButton button;
+
         xcb->last_timestamp = bre->time;
-        nk_bindings_seat_handle_button ( xcb->bindings_seat, bre->detail, NK_BINDINGS_BUTTON_STATE_RELEASE, bre->time );
+        if ( x11_button_to_nk_bindings_button ( bre->detail, &button ) )
+            nk_bindings_seat_handle_button ( xcb->bindings_seat, NULL, button, NK_BINDINGS_BUTTON_STATE_RELEASE, bre->time );
         if ( config.click_to_exit == TRUE ) {
             if ( !xcb->mouse_seen ) {
                 rofi_view_temp_click_to_exit ( state, bre->event );
@@ -722,7 +796,7 @@ static void main_loop_x11_event_handler_view ( xcb_generic_event_t *event )
             for ( gint8 bi = 0; bi < 7; ++bi ) {
                 if ( kne->keys[by] & ( 1 << bi ) ) {
                     // X11 keycodes starts at 8
-                    nk_bindings_seat_handle_key ( xcb->bindings_seat, ( 8 * by + bi ) + 8, NK_BINDINGS_KEY_STATE_PRESSED );
+                    nk_bindings_seat_handle_key ( xcb->bindings_seat, NULL, ( 8 * by + bi ) + 8, NK_BINDINGS_KEY_STATE_PRESSED );
                 }
             }
         }
@@ -734,7 +808,7 @@ static void main_loop_x11_event_handler_view ( xcb_generic_event_t *event )
         gchar                 *text;
 
         xcb->last_timestamp = xkpe->time;
-        text                = nk_bindings_seat_handle_key_with_modmask ( xcb->bindings_seat, xkpe->state, xkpe->detail, NK_BINDINGS_KEY_STATE_PRESS );
+        text                = nk_bindings_seat_handle_key_with_modmask ( xcb->bindings_seat, NULL, xkpe->state, xkpe->detail, NK_BINDINGS_KEY_STATE_PRESS );
         if ( text != NULL ) {
             rofi_view_handle_text ( state, text );
         }
@@ -744,7 +818,7 @@ static void main_loop_x11_event_handler_view ( xcb_generic_event_t *event )
     {
         xcb_key_release_event_t *xkre = (xcb_key_release_event_t *) event;
         xcb->last_timestamp = xkre->time;
-        nk_bindings_seat_handle_key ( xcb->bindings_seat, xkre->detail, NK_BINDINGS_KEY_STATE_RELEASE );
+        nk_bindings_seat_handle_key ( xcb->bindings_seat, NULL, xkre->detail, NK_BINDINGS_KEY_STATE_RELEASE );
         break;
     }
     default:
@@ -783,7 +857,7 @@ static gboolean main_loop_x11_event_handler ( xcb_generic_event_t *ev, G_GNUC_UN
         case XCB_XKB_STATE_NOTIFY:
         {
             xcb_xkb_state_notify_event_t *ksne = (xcb_xkb_state_notify_event_t *) ev;
-            nk_bindings_seat_update_mask ( xcb->bindings_seat,
+            nk_bindings_seat_update_mask ( xcb->bindings_seat, NULL,
                                            ksne->baseMods,
                                            ksne->latchedMods,
                                            ksne->lockedMods,
@@ -912,7 +986,7 @@ static void x11_helper_discover_window_manager ( void )
             if ( wtitle.strings_len > 0 ) {
                 g_debug ( "Found window manager: %s", wtitle.strings );
                 if ( g_strcmp0 ( wtitle.strings, "i3" ) == 0 ) {
-                    current_window_manager = WM_DO_NOT_CHANGE_CURRENT_DESKTOP|WM_PANGO_WORKSPACE_NAMES;
+                    current_window_manager = WM_DO_NOT_CHANGE_CURRENT_DESKTOP | WM_PANGO_WORKSPACE_NAMES;
                 }
             }
             xcb_ewmh_get_utf8_strings_reply_wipe ( &wtitle );
