@@ -43,15 +43,20 @@
 #include <xkbcommon/xkbcommon-compose.h>
 #endif /* NK_XKBCOMMON_HAS_COMPOSE */
 
+#include "nkutils-gtk-settings.h"
+#include "nkutils-xdg-de.h"
 #include "nkutils-enum.h"
 #include "nkutils-bindings.h"
 
+#define NK_BINDINGS_DEFAULT_DOUBLE_CLICK_DELAY 200
+
 #define NK_BINDINGS_MAX_ALIASES 4 /* For Alt */
 
 #define _NK_VALUE_TO_BINDING(mask, value) ((guint64)((((guint64)(mask)) << 32) | (guint32)(value)))
 #define NK_KEYCODE_TO_BINDING(mask, keycode) _NK_VALUE_TO_BINDING(mask, keycode)
 #define NK_KEYSYM_TO_BINDING(mask, keysym) _NK_VALUE_TO_BINDING(mask, keysym)
 #define NK_BUTTON_TO_BINDING(mask, button) _NK_VALUE_TO_BINDING(mask, button)
+#define NK_SCROLL_TO_BINDING(mask, axis, step) _NK_VALUE_TO_BINDING(mask, (axis << 1) + ( (step < 0) ? 0 : 1 ))
 
 struct _NkBindings {
     guint64 double_click_delay;
@@ -81,6 +86,7 @@ typedef struct {
     GHashTable *keycodes;
     GHashTable *keysyms;
     GHashTable *buttons;
+    GHashTable *scroll;
 } NkBindingsScope;
 
 typedef struct {
@@ -137,6 +143,21 @@ static const gchar * const _nk_bindings_modifiers_names[] = {
     [NK_BINDINGS_MODIFIER_HYPER   + NK_BINDINGS_NUM_MODIFIERS * 3] = "",
 };
 
+static const gchar * const _nk_bindings_mouse_button_names[] = {
+    [NK_BINDINGS_MOUSE_BUTTON_PRIMARY] = "Primary",
+    [NK_BINDINGS_MOUSE_BUTTON_SECONDARY] = "Secondary",
+    [NK_BINDINGS_MOUSE_BUTTON_MIDDLE] = "Middle",
+    [NK_BINDINGS_MOUSE_BUTTON_BACK] = "Back",
+    [NK_BINDINGS_MOUSE_BUTTON_FORWARD] = "Forward",
+};
+
+static const gchar * const _nk_bindings_scroll_names[] = {
+    [NK_SCROLL_TO_BINDING(0, NK_BINDINGS_SCROLL_AXIS_VERTICAL, -1)] = "Up",
+    [NK_SCROLL_TO_BINDING(0, NK_BINDINGS_SCROLL_AXIS_VERTICAL,  1)] = "Down",
+    [NK_SCROLL_TO_BINDING(0, NK_BINDINGS_SCROLL_AXIS_HORIZONTAL, -1)] = "Left",
+    [NK_SCROLL_TO_BINDING(0, NK_BINDINGS_SCROLL_AXIS_HORIZONTAL,  1)] = "Right",
+};
+
 GQuark
 nk_bindings_error(void)
 {
@@ -151,18 +172,55 @@ _nk_bindings_scope_free(gpointer data)
     g_hash_table_unref(scope->keycodes);
     g_hash_table_unref(scope->keysyms);
     g_hash_table_unref(scope->buttons);
+    g_hash_table_unref(scope->scroll);
 
     g_slice_free(NkBindingsScope, scope);
 }
 
 NkBindings *
-nk_bindings_new(void)
+nk_bindings_new(guint64 double_click_delay)
 {
     NkBindings *self;
-
     self = g_new0(NkBindings, 1);
 
-    self->double_click_delay = 200;
+    switch ( nk_xdg_de_detect() )
+    {
+    case NK_XDG_DE_NONE:
+    case NK_XDG_DE_GNOME:
+    case NK_XDG_DE_I3:
+        /* Just use the GTK+ settings fallback */
+    break;
+    case NK_XDG_DE_KDE:
+    {
+        gchar *path;
+        GKeyFile *settings;
+
+        path = g_build_filename(g_get_user_config_dir(), "kcminputrc", NULL);
+        settings = g_key_file_new();
+
+        if ( g_key_file_load_from_file(settings, path, G_KEY_FILE_NONE, NULL) )
+            self->double_click_delay = g_key_file_get_uint64(settings, "KDE", "DoubleClickInterval", NULL);
+        g_key_file_free(settings);
+        g_free(path);
+    }
+    break;
+    }
+
+    /* Always fallback to the GTK+ settings */
+    const gchar *gtk_settings_keys[NK_GTK_SETTINGS_NUM_VERSION] = {
+        "gtk-double-click-time",
+        "gtk-double-click-time",
+    };
+    if ( self->double_click_delay < 1 )
+        nk_gtk_settings_get_uint64(&self->double_click_delay, gtk_settings_keys);
+
+    /* If nothing better, use the application value */
+    if ( self->double_click_delay < 1 )
+        self->double_click_delay = double_click_delay;
+
+    /* If nothing better, use the application value */
+    if ( self->double_click_delay < 1 )
+        self->double_click_delay = NK_BINDINGS_DEFAULT_DOUBLE_CLICK_DELAY;
 
     return self;
 }
@@ -176,7 +234,7 @@ nk_bindings_free(NkBindings *self)
 
     g_list_free_full(self->seats, _nk_bindings_seat_free);
 
-    g_list_free_full(self->scopes, _nk_bindings_scope_free);
+    nk_bindings_reset_bindings(self);
 
     g_free(self);
 }
@@ -233,6 +291,7 @@ _nk_bindings_get_scope(NkBindings *self, guint64 scope_id)
         scope->keycodes = g_hash_table_new_full(g_int64_hash, g_int64_equal, NULL, _nk_bindings_binding_free);
         scope->keysyms = g_hash_table_new_full(g_int64_hash, g_int64_equal, NULL, _nk_bindings_binding_free);
         scope->buttons = g_hash_table_new_full(g_int64_hash, g_int64_equal, NULL, _nk_bindings_binding_mouse_free);
+        scope->scroll = g_hash_table_new_full(g_int64_hash, g_int64_equal, NULL, _nk_bindings_binding_free);
         self->scopes = g_list_insert_sorted(self->scopes, scope, _nk_bindings_scope_compare);
     }
 
@@ -264,9 +323,6 @@ nk_bindings_add_binding(NkBindings *self, guint64 scope_id, const gchar *string,
     xkb_mod_mask_t mask = 0;
     xkb_keysym_t last_keysym = XKB_KEY_NoSymbol;
     xkb_keysym_t keysym = XKB_KEY_NoSymbol;
-    xkb_keycode_t keycode = XKB_KEYCODE_INVALID;
-    guint32 button = 0;
-    gboolean double_click = FALSE;
 
     const gchar *w = string;
 
@@ -372,15 +428,21 @@ nk_bindings_add_binding(NkBindings *self, guint64 scope_id, const gchar *string,
             return FALSE;
         }
     }
+
+    NkBindingsScope *scope;
+    scope = _nk_bindings_get_scope(self, scope_id);
+
+    NkBindingsBinding *binding = NULL;
     if ( s < e )
     {
         if ( g_utf8_get_char(s) == '[' )
         {
-            guint64 code;
+            s = g_utf8_next_char(s);
+
+            guint64 keycode;
             gchar *ce;
             errno = 0;
-            s = g_utf8_next_char(s);
-            code = g_ascii_strtoull(s, &ce, 10);
+            keycode = g_ascii_strtoull(s, &ce, 10);
             if ( s == ce )
             {
                 g_set_error(error, NK_BINDINGS_ERROR, NK_BINDINGS_ERROR_SYNTAX, "Syntax error in binding '%s': could not parse keycode", string);
@@ -391,75 +453,104 @@ nk_bindings_add_binding(NkBindings *self, guint64 scope_id, const gchar *string,
                 g_set_error(error, NK_BINDINGS_ERROR, NK_BINDINGS_ERROR_SYNTAX, "Syntax error in binding '%s': keycode must be the end of the binding string, enclosed in squared brackets '[]'", string);
                 return FALSE;
             }
-            else if ( ! xkb_keycode_is_legal_ext(code) )
+            else if ( ! xkb_keycode_is_legal_ext(keycode) )
             {
-                g_set_error(error, NK_BINDINGS_ERROR, NK_BINDINGS_ERROR_SYNTAX, "Syntax error in binding '%s': wrong keycode value %"G_GINT64_MODIFIER"u", string, code);
+                g_set_error(error, NK_BINDINGS_ERROR, NK_BINDINGS_ERROR_SYNTAX, "Syntax error in binding '%s': wrong keycode value %"G_GINT64_MODIFIER"u", string, keycode);
                 return FALSE;
             }
-            keycode = code;
+
+            guint64 value = NK_KEYCODE_TO_BINDING(mask, keycode);
+            binding = g_hash_table_lookup(scope->keycodes, &value);
+            if ( binding == NULL )
+            {
+                binding = g_slice_new0(NkBindingsBinding);
+                binding->binding = value;
+                g_hash_table_insert(scope->keycodes, &binding->binding, binding);
+            }
         }
         else if ( g_ascii_strncasecmp(s, "Mouse", strlen("Mouse")) == 0 )
         {
             s += strlen("Mouse");
+
+            gboolean double_click = FALSE;
             if ( g_unichar_toupper(g_utf8_get_char(s)) == 'D' )
             {
                 s = g_utf8_next_char(s);
                 double_click = TRUE;
             }
-            guint64 code;
-            gchar *ce;
-            errno = 0;
-            code = g_ascii_strtoull(s, &ce, 10);
-            if ( s == ce )
+
+            guint64 button;
+            if ( g_ascii_strncasecmp(s, "Extra", strlen("Extra")) == 0 )
             {
-                g_set_error(error, NK_BINDINGS_ERROR, NK_BINDINGS_ERROR_SYNTAX, "Syntax error in binding '%s': could not parse mouse button number", string);
+                s += strlen("Extra");
+                gchar *ce;
+                errno = 0;
+                button = g_ascii_strtoull(s, &ce, 10);
+                if ( s == ce )
+                {
+                    g_set_error(error, NK_BINDINGS_ERROR, NK_BINDINGS_ERROR_SYNTAX, "Syntax error in binding '%s': could not parse mouse extra button number", string);
+                    return FALSE;
+                }
+                else if ( e != ce )
+                {
+                    g_set_error(error, NK_BINDINGS_ERROR, NK_BINDINGS_ERROR_SYNTAX, "Syntax error in binding '%s': mouse extra button must be the end of the binding string", string);
+                    return FALSE;
+                }
+                button += NK_BINDINGS_MOUSE_BUTTON_EXTRA;
+            }
+            else if ( ! nk_enum_parse(s, _nk_bindings_mouse_button_names, G_N_ELEMENTS(_nk_bindings_mouse_button_names), TRUE, FALSE, &button) )
+            {
+                g_set_error(error, NK_BINDINGS_ERROR, NK_BINDINGS_ERROR_SYNTAX, "Syntax error in binding '%s': unknown mouse button %s", string, s);
                 return FALSE;
             }
-            else if ( e != ce )
+
+            guint64 value = NK_BUTTON_TO_BINDING(mask, button);
+            NkBindingsBindingMouse *mouse_binding;
+            mouse_binding = g_hash_table_lookup(scope->buttons, &value);
+            if ( mouse_binding == NULL )
             {
-                g_set_error(error, NK_BINDINGS_ERROR, NK_BINDINGS_ERROR_SYNTAX, "Syntax error in binding '%s': keycode must be the end of the binding string", string);
+                mouse_binding = g_slice_new0(NkBindingsBindingMouse);
+                mouse_binding->click.binding = value;
+                mouse_binding->dclick.binding = value;
+                g_hash_table_insert(scope->buttons, &mouse_binding->click.binding, mouse_binding);
+            }
+
+            if ( double_click )
+                binding = &mouse_binding->dclick;
+            else
+                binding = &mouse_binding->click;
+        }
+        else if ( g_ascii_strncasecmp(s, "Scroll", strlen("Scroll")) == 0 )
+        {
+            s += strlen("Scroll");
+
+            if ( on_release )
+            {
+                g_set_error(error, NK_BINDINGS_ERROR, NK_BINDINGS_ERROR_SYNTAX, "Syntax error in binding '%s': scroll action cannot be bound on release", string);
                 return FALSE;
             }
-            button = code;
+
+            guint64 scroll;
+            if ( ! nk_enum_parse(s, _nk_bindings_scroll_names, G_N_ELEMENTS(_nk_bindings_scroll_names), TRUE, FALSE, &scroll) )
+            {
+                g_set_error(error, NK_BINDINGS_ERROR, NK_BINDINGS_ERROR_SYNTAX, "Syntax error in binding '%s': unknown scroll direction %s", string, s);
+                return FALSE;
+            }
+
+            guint64 value = _NK_VALUE_TO_BINDING(mask, scroll);
+            binding = g_hash_table_lookup(scope->scroll, &value);
+            if ( binding == NULL )
+            {
+                binding = g_slice_new0(NkBindingsBinding);
+                binding->binding = value;
+                g_hash_table_insert(scope->scroll, &binding->binding, binding);
+            }
         }
         else
             keysym = xkb_keysym_from_name(s, XKB_KEYSYM_NO_FLAGS);
     }
 
-    NkBindingsScope *scope;
-    scope = _nk_bindings_get_scope(self, scope_id);
-
-    NkBindingsBinding *binding = NULL;
-    if ( button != 0 )
-    {
-        guint64 value = NK_BUTTON_TO_BINDING(mask, button);
-        NkBindingsBindingMouse *mouse_binding;
-        mouse_binding = g_hash_table_lookup(scope->buttons, &value);
-        if ( mouse_binding == NULL )
-        {
-            mouse_binding = g_slice_new0(NkBindingsBindingMouse);
-            mouse_binding->click.binding = value;
-            mouse_binding->dclick.binding = value;
-            g_hash_table_insert(scope->buttons, &mouse_binding->click.binding, mouse_binding);
-        }
-
-        if ( double_click )
-            binding = &mouse_binding->dclick;
-        else
-            binding = &mouse_binding->click;
-    }
-    else if ( keycode != XKB_KEYCODE_INVALID )
-    {
-        guint64 value = NK_KEYCODE_TO_BINDING(mask, keycode);
-        binding = g_hash_table_lookup(scope->keycodes, &value);
-        if ( binding == NULL )
-        {
-            binding = g_slice_new0(NkBindingsBinding);
-            binding->binding = value;
-            g_hash_table_insert(scope->keycodes, &binding->binding, binding);
-        }
-    }
-    else
+    if ( binding == NULL )
     {
         if ( keysym == XKB_KEY_NoSymbol )
         {
@@ -497,8 +588,15 @@ nk_bindings_add_binding(NkBindings *self, guint64 scope_id, const gchar *string,
     return TRUE;
 }
 
+void
+nk_bindings_reset_bindings(NkBindings *self)
+{
+    g_list_free_full(self->scopes, _nk_bindings_scope_free);
+    self->scopes = NULL;
+}
+
 static gboolean
-_nk_bindings_seat_binding_trigger(NkBindingsSeat *self, NkBindingsBinding *binding, gboolean trigger)
+_nk_bindings_seat_binding_trigger(NkBindingsSeat *self, NkBindingsBinding *binding, gpointer target, gboolean trigger)
 {
     if ( binding == NULL )
         return FALSE;
@@ -507,7 +605,7 @@ _nk_bindings_seat_binding_trigger(NkBindingsSeat *self, NkBindingsBinding *bindi
     gboolean has_press = ( binding->press.base.callback != NULL );
     gboolean on_release_waiting = ( g_list_find(self->on_release, binding) != NULL );
     if ( trigger && has_press )
-        handled = binding->press.base.callback(binding->scope, binding->press.base.user_data);
+        handled = binding->press.base.callback(binding->scope, target, binding->press.base.user_data);
     if ( ( ! on_release_waiting ) && ( binding->release.base.callback != NULL ) && ( handled || ( ! trigger ) || ( ! has_press ) ) )
     {
         self->on_release = g_list_prepend(self->on_release, binding);
@@ -518,15 +616,15 @@ _nk_bindings_seat_binding_trigger(NkBindingsSeat *self, NkBindingsBinding *bindi
 }
 
 static gboolean
-_nk_bindings_try_key_bindings(NkBindings *self, NkBindingsSeat *seat, guint64 keycode, guint64 keysym, gboolean trigger)
+_nk_bindings_try_key_bindings(NkBindings *self, NkBindingsSeat *seat, gpointer target, guint64 keycode, guint64 keysym, gboolean trigger)
 {
     GList *scope_;
     for ( scope_ = self->scopes ; scope_ != NULL ; scope_ = g_list_next(scope_) )
     {
         NkBindingsScope *scope = scope_->data;
-        if ( _nk_bindings_seat_binding_trigger(seat, g_hash_table_lookup(scope->keycodes, &keycode), trigger) )
+        if ( _nk_bindings_seat_binding_trigger(seat, g_hash_table_lookup(scope->keycodes, &keycode), target, trigger) )
             return TRUE;
-        if ( ( keysym != XKB_KEY_NoSymbol ) && _nk_bindings_seat_binding_trigger(seat, g_hash_table_lookup(scope->keysyms, &keysym), trigger) )
+        if ( ( keysym != XKB_KEY_NoSymbol ) && _nk_bindings_seat_binding_trigger(seat, g_hash_table_lookup(scope->keysyms, &keysym), target, trigger) )
             return TRUE;
     }
 
@@ -534,7 +632,7 @@ _nk_bindings_try_key_bindings(NkBindings *self, NkBindingsSeat *seat, guint64 ke
 }
 
 static gboolean
-_nk_bindings_try_button_bindings(NkBindings *self, NkBindingsSeat *seat, guint64 **button, guint64 time)
+_nk_bindings_try_button_bindings(NkBindings *self, NkBindingsSeat *seat, gpointer target, guint64 **button, guint64 time)
 {
     NkBindingsBindingMouse *mouse_binding = NULL;
     GList *scope_;
@@ -548,13 +646,13 @@ _nk_bindings_try_button_bindings(NkBindings *self, NkBindingsSeat *seat, guint64
 
         if ( time < self->double_click_delay )
         {
-            if ( _nk_bindings_seat_binding_trigger(seat, &mouse_binding->dclick, TRUE) )
+            if ( _nk_bindings_seat_binding_trigger(seat, &mouse_binding->dclick, target, TRUE) )
             {
                 *button = &mouse_binding->click.binding;
                 return TRUE;
             }
         }
-        if ( _nk_bindings_seat_binding_trigger(seat, &mouse_binding->click, TRUE) )
+        if ( _nk_bindings_seat_binding_trigger(seat, &mouse_binding->click, target, TRUE) )
         {
             *button = &mouse_binding->click.binding;
             return TRUE;
@@ -563,6 +661,23 @@ _nk_bindings_try_button_bindings(NkBindings *self, NkBindingsSeat *seat, guint64
     return FALSE;
 }
 
+static NkBindingsBinding *
+_nk_bindings_try_scroll_bindings(NkBindings *self, NkBindingsSeat *seat, gpointer target, guint64 scroll)
+{
+    GList *scope_;
+    for ( scope_ = self->scopes ; scope_ != NULL ; scope_ = g_list_next(scope_) )
+    {
+        NkBindingsScope *scope = scope_->data;
+        NkBindingsBinding *binding;
+
+        binding = g_hash_table_lookup(scope->scroll, &scroll);
+        if ( _nk_bindings_seat_binding_trigger(seat, binding, target, TRUE) )
+            return binding;
+    }
+
+    return NULL;
+}
+
 NkBindingsSeat *
 nk_bindings_seat_new(NkBindings *bindings, enum xkb_context_flags flags)
 {
@@ -587,13 +702,13 @@ nk_bindings_seat_new(NkBindings *bindings, enum xkb_context_flags flags)
     return self;
 }
 
-static void _nk_bindings_seat_free_on_release(NkBindingsSeat *self, gboolean trigger);
+static void _nk_bindings_seat_free_on_release(NkBindingsSeat *self, gpointer target, gboolean trigger);
 static void
 _nk_bindings_seat_free(gpointer data)
 {
     NkBindingsSeat *self = data;
 
-    _nk_bindings_seat_free_on_release(self, FALSE);
+    _nk_bindings_seat_free_on_release(self, NULL, FALSE);
 
     xkb_keymap_unref(self->keymap);
     xkb_state_unref(self->state);
@@ -634,10 +749,22 @@ void
 nk_bindings_seat_update_keymap(NkBindingsSeat *self, struct xkb_keymap *keymap, struct xkb_state *state)
 {
     g_return_if_fail(self != NULL);
+    g_return_if_fail((keymap != NULL && state != NULL) || (keymap == NULL && state == NULL));
 
     xkb_keymap_unref(self->keymap);
     xkb_state_unref(self->state);
 
+    if ( keymap == NULL )
+    {
+        nk_bindings_seat_reset(self);
+        self->keymap = NULL;
+        self->state = NULL;
+        gsize i;
+        for ( i = 0 ; i < NK_BINDINGS_NUM_MODIFIERS ; ++i )
+            self->modifiers[i][0] = XKB_MOD_INVALID;
+        return;
+    }
+
     self->keymap = xkb_keymap_ref(keymap);
     self->state = xkb_state_ref(state);
 
@@ -662,6 +789,9 @@ _nk_bindings_seat_get_modifiers_masks(NkBindingsSeat *self, xkb_keycode_t key, x
     *effective = 0;
     *not_consumed = 0;
 
+    if ( self->state == NULL )
+        return;
+
     NkBindingsModifiers mod;
     xkb_mod_index_t *i;
     for ( mod = 0 ; mod < NK_BINDINGS_NUM_MODIFIERS ; ++mod )
@@ -684,7 +814,7 @@ _nk_bindings_seat_get_modifiers_masks(NkBindingsSeat *self, xkb_keycode_t key, x
 }
 
 static void
-_nk_bindings_seat_free_on_release(NkBindingsSeat *self, gboolean trigger)
+_nk_bindings_seat_free_on_release(NkBindingsSeat *self, gpointer target, gboolean trigger)
 {
     if ( self->on_release == NULL )
         return;
@@ -694,16 +824,18 @@ _nk_bindings_seat_free_on_release(NkBindingsSeat *self, gboolean trigger)
     {
         NkBindingsBinding *binding = link->data;
         if ( trigger )
-            binding->release.base.callback(binding->scope, binding->release.base.user_data);
+            binding->release.base.callback(binding->scope, target, binding->release.base.user_data);
         g_list_free_1(link);
     }
     self->on_release = NULL;
 }
 
 gchar *
-nk_bindings_seat_handle_key(NkBindingsSeat *self, xkb_keycode_t keycode, NkBindingsKeyState state)
+nk_bindings_seat_handle_key(NkBindingsSeat *self, gpointer target, xkb_keycode_t keycode, NkBindingsKeyState state)
 {
     g_return_val_if_fail(self != NULL, NULL);
+    g_return_val_if_fail(self->keymap != NULL, NULL);
+    g_return_val_if_fail(self->state != NULL, NULL);
 
     xkb_keysym_t keysym;
     gchar *tmp = NULL;
@@ -714,7 +846,7 @@ nk_bindings_seat_handle_key(NkBindingsSeat *self, xkb_keycode_t keycode, NkBindi
         xkb_mod_mask_t dummy, mask;
         _nk_bindings_seat_get_modifiers_masks(self, 0, &dummy, &mask);
         if ( mask == 0 )
-            _nk_bindings_seat_free_on_release(self, TRUE);
+            _nk_bindings_seat_free_on_release(self, target, TRUE);
         return NULL;
     }
 
@@ -749,7 +881,7 @@ nk_bindings_seat_handle_key(NkBindingsSeat *self, xkb_keycode_t keycode, NkBindi
     xkb_mod_mask_t effective, not_consumed;
     _nk_bindings_seat_get_modifiers_masks(self, keycode, &effective, &not_consumed);
 
-    if ( _nk_bindings_try_key_bindings(self->bindings, self, NK_KEYCODE_TO_BINDING(effective, keycode), NK_KEYSYM_TO_BINDING(not_consumed, keysym), regular_press) )
+    if ( _nk_bindings_try_key_bindings(self->bindings, self, target, NK_KEYCODE_TO_BINDING(effective, keycode), NK_KEYSYM_TO_BINDING(not_consumed, keysym), regular_press) )
         return NULL;
 
     if ( ! regular_press )
@@ -769,8 +901,12 @@ nk_bindings_seat_handle_key(NkBindingsSeat *self, xkb_keycode_t keycode, NkBindi
 }
 
 gchar *
-nk_bindings_seat_handle_key_with_modmask(NkBindingsSeat *self, xkb_mod_mask_t modmask, xkb_keycode_t keycode, NkBindingsKeyState state)
+nk_bindings_seat_handle_key_with_modmask(NkBindingsSeat *self, gpointer target, xkb_mod_mask_t modmask, xkb_keycode_t keycode, NkBindingsKeyState state)
 {
+    g_return_val_if_fail(self != NULL, NULL);
+    g_return_val_if_fail(self->keymap != NULL, NULL);
+    g_return_val_if_fail(self->state != NULL, NULL);
+
     xkb_mod_mask_t effective_mods, depressed_mods, latched_mods, locked_mods;
     xkb_layout_index_t depressed_layout, latched_layout, locked_layout;
     gchar *ret;
@@ -784,17 +920,17 @@ nk_bindings_seat_handle_key_with_modmask(NkBindingsSeat *self, xkb_mod_mask_t mo
     locked_layout = xkb_state_serialize_layout(self->state, XKB_STATE_LAYOUT_LOCKED);
 
     if ( effective_mods == modmask )
-        return nk_bindings_seat_handle_key(self, keycode, state);
+        return nk_bindings_seat_handle_key(self, target, keycode, state);
 
-    xkb_state_update_mask(self->state, modmask, 0, 0, 0, 0, locked_layout);
-    ret = nk_bindings_seat_handle_key(self, keycode, state);
+    xkb_state_update_mask(self->state, modmask, 0, 0, depressed_layout, latched_layout, locked_layout);
+    ret = nk_bindings_seat_handle_key(self, target, keycode, state);
     xkb_state_update_mask(self->state, depressed_mods, latched_mods, locked_mods, depressed_layout, latched_layout, locked_layout);
 
     return ret;
 }
 
 gboolean
-nk_bindings_seat_handle_button(NkBindingsSeat *self, guint32 button, NkBindingsButtonState state, guint64 timestamp)
+nk_bindings_seat_handle_button(NkBindingsSeat *self, gpointer target, NkBindingsMouseButton button, NkBindingsButtonState state, guint64 timestamp)
 {
     g_return_val_if_fail(self != NULL, FALSE);
 
@@ -805,7 +941,7 @@ nk_bindings_seat_handle_button(NkBindingsSeat *self, guint32 button, NkBindingsB
     if ( state == NK_BINDINGS_BUTTON_STATE_RELEASE )
     {
         if ( mask == 0 )
-            _nk_bindings_seat_free_on_release(self, TRUE);
+            _nk_bindings_seat_free_on_release(self, target, TRUE);
         return TRUE;
     }
 
@@ -815,7 +951,7 @@ nk_bindings_seat_handle_button(NkBindingsSeat *self, guint32 button, NkBindingsB
     last_timestamp = g_hash_table_lookup(self->last_timestamps, &binding);
     if ( last_timestamp != NULL )
         time -= *last_timestamp;
-    if ( ! _nk_bindings_try_button_bindings(self->bindings, self, &binding_, time) )
+    if ( ! _nk_bindings_try_button_bindings(self->bindings, self, target, &binding_, time) )
         return FALSE;
     if ( last_timestamp == NULL )
         g_hash_table_insert(self->last_timestamps, binding_, g_memdup(&timestamp, sizeof(guint64)));
@@ -824,10 +960,35 @@ nk_bindings_seat_handle_button(NkBindingsSeat *self, guint32 button, NkBindingsB
     return TRUE;
 }
 
+gboolean
+nk_bindings_seat_handle_scroll(NkBindingsSeat *self, gpointer target, NkBindingsScrollAxis axis, gint32 step)
+{
+    g_return_val_if_fail(self != NULL, FALSE);
+    g_return_val_if_fail(step != 0, FALSE);
+
+    xkb_mod_mask_t dummy, mask;
+
+    _nk_bindings_seat_get_modifiers_masks(self, 0, &dummy, &mask);
+
+    guint64 binding_ = NK_SCROLL_TO_BINDING(mask, axis, step);
+
+    NkBindingsBinding *binding;
+    binding = _nk_bindings_try_scroll_bindings(self->bindings, self, target, binding_);
+
+    if ( binding == NULL )
+        return FALSE;
+
+    for ( step = ABS(step) ; step > 1 ; --step )
+        _nk_bindings_seat_binding_trigger(self, binding, target, TRUE);
+    return TRUE;
+}
+
 void
-nk_bindings_seat_update_mask(NkBindingsSeat *self, xkb_mod_mask_t depressed_mods, xkb_mod_mask_t latched_mods, xkb_mod_mask_t locked_mods, xkb_layout_index_t depressed_layout, xkb_layout_index_t latched_layout, xkb_layout_index_t locked_layout)
+nk_bindings_seat_update_mask(NkBindingsSeat *self, gpointer target, xkb_mod_mask_t depressed_mods, xkb_mod_mask_t latched_mods, xkb_mod_mask_t locked_mods, xkb_layout_index_t depressed_layout, xkb_layout_index_t latched_layout, xkb_layout_index_t locked_layout)
 {
     g_return_if_fail(self != NULL);
+    g_return_if_fail(self->state != NULL);
+    g_return_if_fail(self->keymap != NULL);
 
     enum xkb_state_component changed;
 
@@ -838,7 +999,7 @@ nk_bindings_seat_update_mask(NkBindingsSeat *self, xkb_mod_mask_t depressed_mods
         xkb_mod_mask_t dummy, mask;
         _nk_bindings_seat_get_modifiers_masks(self, 0, &dummy, &mask);
         if ( mask == 0 )
-            _nk_bindings_seat_free_on_release(self, TRUE);
+            _nk_bindings_seat_free_on_release(self, target, TRUE);
     }
 }
 
@@ -847,5 +1008,5 @@ nk_bindings_seat_reset(NkBindingsSeat *self)
 {
     g_return_if_fail(self != NULL);
 
-    _nk_bindings_seat_free_on_release(self, FALSE);
+    _nk_bindings_seat_free_on_release(self, NULL, FALSE);
 }
