@@ -40,6 +40,8 @@
 #include <gio/gio.h>
 
 #include "nkutils-enum.h"
+#include "nkutils-gtk-settings.h"
+#include "nkutils-xdg-de.h"
 #include "nkutils-xdg-theme.h"
 
 typedef enum {
@@ -51,30 +53,6 @@ typedef enum {
 #define NK_XDG_THEME_ICON_FALLBACK_THEME "hicolor"
 #define NK_XDG_THEME_SOUND_FALLBACK_THEME "freedesktop"
 
-typedef enum {
-    DE_NONE = 0,
-    DE_GNOME,
-    DE_KDE,
-} NkXdgThemeDE;
-
-static const gchar * const _nk_xdg_theme_de_session_desktop_names[] = {
-    [DE_NONE] = "generic",
-    [DE_GNOME] = "GNOME",
-    [DE_KDE] = "KDE",
-};
-
-static const gchar * const _nk_xdg_theme_de_current_session_names[] = {
-    [DE_NONE] = "X-Generic",
-    [DE_GNOME] = "GNOME",
-    [DE_KDE] = "KDE",
-};
-
-static const gchar * const _nk_xdg_theme_de_desktop_session_names[] = {
-    [DE_NONE] = "generic",
-    [DE_GNOME] = "gnome",
-    [DE_KDE] = "kde",
-};
-
 typedef struct {
     const gchar *schema_id;
     const gchar *signal_name;
@@ -103,6 +81,7 @@ typedef struct {
     gchar *de_theme;
     gpointer de_data;
     GDestroyNotify de_notify;
+    gchar *gtk_theme;
 } NkXdgThemeTypeContext;
 
 struct _NkXdgThemeContext {
@@ -237,55 +216,6 @@ static const gchar *_nk_xdg_theme_sound_extensions[] = {
     NULL
 };
 
-static NkXdgThemeDE
-_nk_xdg_theme_de_detect(void)
-{
-    static NkXdgThemeDE _nk_xdg_theme_de = DE_NONE - 1;
-    if ( _nk_xdg_theme_de != (NkXdgThemeDE) ( DE_NONE - 1 ) )
-        return _nk_xdg_theme_de;
-
-    const gchar *var;
-    guint64 value;
-
-    var = g_getenv("XDG_SESSION_DESKTOP");
-    if ( ( var != NULL ) && nk_enum_parse(var, _nk_xdg_theme_de_session_desktop_names, G_N_ELEMENTS(_nk_xdg_theme_de_session_desktop_names), TRUE, TRUE, &value) )
-    {
-        _nk_xdg_theme_de = value;
-        return _nk_xdg_theme_de;
-    }
-
-    var = g_getenv("XDG_CURRENT_DESKTOP");
-    if ( ( var != NULL ) && nk_enum_parse(var, _nk_xdg_theme_de_current_session_names, G_N_ELEMENTS(_nk_xdg_theme_de_current_session_names), FALSE, TRUE, &value) )
-    {
-        _nk_xdg_theme_de = value;
-        return _nk_xdg_theme_de;
-    }
-
-    var = g_getenv("GNOME_DESKTOP_SESSION_ID");
-    if ( ( var != NULL ) && ( *var != '\0' ) )
-    {
-        _nk_xdg_theme_de = DE_GNOME;
-        return _nk_xdg_theme_de;
-    }
-
-    var = g_getenv("KDE_FULL_SESSION");
-    if ( ( var != NULL ) && ( *var != '\0' ) )
-    {
-        _nk_xdg_theme_de = DE_KDE;
-        return _nk_xdg_theme_de;
-    }
-
-    var = g_getenv("DESKTOP_SESSION");
-    if ( ( var != NULL ) && nk_enum_parse(var, _nk_xdg_theme_de_desktop_session_names, G_N_ELEMENTS(_nk_xdg_theme_de_desktop_session_names), FALSE, FALSE, &value) )
-    {
-        _nk_xdg_theme_de = value;
-        return _nk_xdg_theme_de;
-    }
-
-    _nk_xdg_theme_de = DE_NONE;
-    return _nk_xdg_theme_de;
-}
-
 static void
 _nk_xdg_theme_de_theme_gsettings_update(NkXdgThemeTypeContext *self, G_GNUC_UNUSED gchar *key, GSettings *settings)
 {
@@ -296,11 +226,12 @@ _nk_xdg_theme_de_theme_gsettings_update(NkXdgThemeTypeContext *self, G_GNUC_UNUS
 static void
 _nk_xdg_theme_de_theme_hook(NkXdgThemeTypeContext *self)
 {
-    switch ( _nk_xdg_theme_de_detect() )
+    switch ( nk_xdg_de_detect() )
     {
-    case DE_NONE:
+    case NK_XDG_DE_NONE:
+    case NK_XDG_DE_I3:
     break;
-    case DE_GNOME:
+    case NK_XDG_DE_GNOME:
     {
         const NkXdgThemeGnomeSettings *settings = &_nk_xdg_theme_gnome_settings[self->type];
         GSettingsSchemaSource *schema_source;
@@ -322,7 +253,7 @@ _nk_xdg_theme_de_theme_hook(NkXdgThemeTypeContext *self)
         self->de_notify = g_object_unref;
     }
     break;
-    case DE_KDE:
+    case NK_XDG_DE_KDE:
         switch ( self->type )
         {
         case TYPE_ICON:
@@ -344,6 +275,12 @@ _nk_xdg_theme_de_theme_hook(NkXdgThemeTypeContext *self)
         }
     break;
     }
+
+    const gchar *gtk_settings_keys[NK_GTK_SETTINGS_NUM_VERSION] = {
+        "gtk-icon-theme-name",
+        "gtk-icon-theme-name",
+    };
+    nk_gtk_settings_get_string(&self->gtk_theme, gtk_settings_keys);
 }
 
 static void
@@ -360,7 +297,7 @@ _nk_xdg_theme_find_dirs(NkXdgThemeTypeContext *self)
     switch ( self->type )
     {
     case TYPE_ICON:
-        ++length; /* pixmaps */
+        length *= 2; /* ~/.icons and pixmaps */
     break;
     case TYPE_SOUND:
     break;
@@ -379,14 +316,29 @@ _nk_xdg_theme_find_dirs(NkXdgThemeTypeContext *self)
 
     try_dir(g_get_user_data_dir());
 
-    const gchar * const *system_dir;
-    for ( system_dir = system_dirs ; *system_dir != NULL ; ++system_dir )
-        try_dir(*system_dir);
     switch ( self->type )
     {
     case TYPE_ICON:
-        if ( g_file_test(G_DIR_SEPARATOR_S "usr" G_DIR_SEPARATOR_S "share" G_DIR_SEPARATOR_S "pixmaps", G_FILE_TEST_IS_DIR) ) \
-            dirs[current++] = g_strdup(G_DIR_SEPARATOR_S "usr" G_DIR_SEPARATOR_S "share" G_DIR_SEPARATOR_S "pixmaps");\
+        subdir = ".icons";
+        try_dir(g_get_home_dir());
+        subdir = _nk_xdg_theme_subdirs[self->type];
+    break;
+    case TYPE_SOUND:
+    break;
+    }
+
+    const gchar * const *system_dir;
+    for ( system_dir = system_dirs ; *system_dir != NULL ; ++system_dir )
+        try_dir(*system_dir);
+
+    switch ( self->type )
+    {
+    case TYPE_ICON:
+        /* Let’s trust system_dirs to contains /usr/share, if it doesn’t,
+         * that probably means it would be useless to check it anyway */
+        subdir = "pixmaps";
+        for ( system_dir = system_dirs ; *system_dir != NULL ; ++system_dir )
+            try_dir(*system_dir);
     break;
     case TYPE_SOUND:
     break;
@@ -746,6 +698,7 @@ nk_xdg_theme_context_free(NkXdgThemeContext *context)
     {
         NkXdgThemeTypeContext *self = &context->types[type];
 
+        g_free(self->gtk_theme);
         if ( self->de_notify != NULL )
             self->de_notify(self->de_data);
         g_free(self->de_theme);
@@ -791,6 +744,13 @@ _nk_xdg_theme_foreach_theme(NkXdgThemeTypeContext *self, const gchar * const *th
             return TRUE;
     }
 
+    if ( self->gtk_theme != NULL )
+    {
+        theme = _nk_xdg_theme_get_theme(self, self->gtk_theme);
+        if ( ( theme != NULL ) && callback(theme, data, ret) )
+            return TRUE;
+    }
+
     for ( theme_name = self->fallback_themes ; *theme_name != NULL ; ++theme_name )
     {
         theme = _nk_xdg_theme_get_theme(self, *theme_name);
